"""
pred_batch = <tf.Tensor 'pred_batch:0' shape=(None,) dtype=string>
sim = <tf.Tensor 'while/StatefulPartitionedCall:0' shape=<unknown> dtype=float32>
pred_batch[i] = <tf.Tensor 'while/strided_slice_3:0' shape=() dtype=string>

"""

import unicodedata

import numpy as np


COST = {
    "DELETE": 1,
    "INSERT": 1,
    "SUBSTITUTE": 2,
}

def lev(pred:str, gt:str, cost:dict=COST) -> float:
    """
    Levenshtein distance
    """
    m = len(pred)
    n = len(gt)
    # Think of distance_matrix as a (m+1, n+1)-matrix
    distance_matrix = dict()

    distance_matrix[0,0] = 0
    for i in range(1,m+1):
        distance_matrix[i,0] = distance_matrix[i-1,0] + cost["DELETE"]
    for j in range(1,n+1):
        distance_matrix[0,j] = distance_matrix[0,j-1] + cost["INSERT"]

    for i in range(1,m+1):
        for j in range(1,n+1):
            candidates = (
                distance_matrix[i-1, j-1] + (0 if pred[i-1] == gt[j-1] else cost["SUBSTITUTE"]),
                distance_matrix[i-1, j] + cost["DELETE"],
                distance_matrix[i, j-1] + cost["INSERT"],
            )
            distance_matrix[i,j] = min(candidates)

    return distance_matrix[m, n]


def norm_lev(
    pred:str,
    gt:str,
    *,
    form:str="NFD",
    cost:dict=COST,
):
    pred = unicodedata.normalize(form, pred)
    gt = unicodedata.normalize(form, gt)
    return lev(pred, gt, cost)


Batch = list[str]

def batch_lev(
    pred_batch: Batch,
    gt_batch: Batch,
    *,
    form:str="NFD",
    cost:dict=COST,
):
    somme = 0
    for pred, gt in zip(pred_batch, gt_batch):
        somme += norm_lev(pred, gt, form=form, cost=cost)
    avg = somme / len(pred_batch)
    return avg


def sim(
    pred:str,
    gt:str,
    *,
    alpha:float=1/3,
    soft:bool=True,
    proportional:bool=True,
) -> float:
    """
    The idea is that we need this function to return an
    accuracy-like score, in the range of [0, 1],
    to measure the similarity between two strings.

    We have provided two distinct definition for soft similarity,
    controlled by the boolean `proportional`
    """
    if soft:
        dist = lev(pred, gt)
        if proportional:
            proportion = dist / (COST["SUBSTITUTE"]*max(len(gt), 1))
            similarity = 1 - min(1, proportion)
        else:
            if dist > len(gt)*COST["SUBSTITUTE"]:
                # That is, when they are completely different strings
                similarity = 0
            else:
                similarity = 1 / (alpha*dist + 1)
    else:
        # hard similarity is binary: same string or not
        similarity = float(pred == gt)
    return similarity


def norm_sim(
    pred:str,
    gt:str,
    *,
    alpha:float=1/3,
    soft:bool=True,
    proportional:bool=True,
    form:str="NFD",
) -> float:
    pred = unicodedata.normalize(form, pred)
    gt = unicodedata.normalize(form, gt)
    return sim(pred, gt, alpha=alpha, soft=soft, proportional=proportional)


def batch_sim(
    pred_batch: Batch,
    gt_batch: Batch,
    *,
    alpha:float=1/3,
    soft:bool=True,
    proportional:bool=True,
    form:str="NFD",
) -> float:
    somme = 0
    for pred, gt in zip(pred_batch, gt_batch):
        somme += norm_sim(pred, gt, alpha=alpha, soft=soft, proportional=proportional, form=form)
    avg = somme / len(pred_batch)
    return avg


def stdout_batch_sim(
    pred_batch: Batch,
    gt_batch: Batch,
    *,
    alpha:float=1/3,
    soft:bool=True,
    proportional:bool=True,
    form:str="NFD",
    verbose:bool=False,
) -> float:
    if verbose:
        for i, (pred, gt) in enumerate(zip(pred_batch, gt_batch)):
            sim_score = norm_sim(pred, gt)
            print(f"({i = })")
            print(f"{sim_score = :.2f}")
            print(f"{pred = }")
            print(f"{gt = }")
            print()
    avg_sim_score = batch_sim(pred_batch, gt_batch)
    print(f"{avg_sim_score = :.2f}")


if __name__ == "__main__":
    #pred = "Tôi ở Sài Gòn được 6 năm rồi."
    ##gt = "Tôi ở Sài Gòn được 7 năm."
    #gt = "Tôi ở đây chưa lâu."
    #print(f"{pred = }")
    #print(f"{gt = }")
    #d = norm_lev(pred, gt, form="NFKD")
    #print(f"(Python land) {d = }")
    #pred = tf.constant(pred)
    #gt = tf.constant(gt)
    #d = tf_byte_lev(pred, gt)
    #print(f"(TF land)     {d = }")
    #print()
    #proportional_similarity = tf_byte_sim(pred, gt)
    #print(f"(TF land)     {proportional_similarity = }")
    #unproportional_similarity = tf_byte_sim(pred, gt, proportional=tf.constant(False))
    #print(f"(TF land)     {unproportional_similarity = }")
    pred_batch = [
        "Tôi ở Sài Gòn được 6 năm rồi.",
        "Em chưa đi Đà Lạt bao giờ.",
        "Tối nay anh có đi year end party không?",
    ]
    gt_batch = [
        "Toi o Sái Gòn dưoc 6 nam rồi.",
        "Em chua di Dà Lát bao giờ!",
        "Tối nay anh có đi xear ant party không?",
    ]
    stdout_batch_sim(pred_batch, gt_batch, verbose=True)



import unicodedata

from tqdm.auto import tqdm
from rapidfuzz.distance import Levenshtein

from vina2vi.util import normalize


def get_avg_sim(
    preds,
    gts,
    *args,
    n_strings: int|None = None,
):
    """
    args
        preds
            a sequence of strings, predictions of some model
        gts
            a sequence of groundtruth strings (i.e. with diacritics)
        n_strings
            the number of strings contained in preds (or gts)
    """
    norm_preds = (normalize(s) for s in preds)
    norm_gts = (normalize(s) for s in gts)
    sims = []
    if n_strings is None and "__len__" in dir(gts):
        n_strings = len(gts)
    for i, (p, g) in tqdm(enumerate(zip(norm_preds, norm_gts)), total=n_strings):
        sims.append(Levenshtein.normalized_similarity(p, g))

    if not "i" in locals():
        print("Cannot compute avg_sim. Maybe you have an empty preds or gts?")
        return None

    # n_instances and n_strings are synonyms in this function
    n_instances = i+1
    avg_sim = sum(sims)/n_instances
    return avg_sim





import string
import unicodedata
from dataclasses import dataclass, field

import unidecode
import torch
from datasets import Dataset, DatasetDict


@dataclass(frozen=True)
class Vietnamese:
    ambiguous_chars = {
        "a": "aáảãàạâấẩẫầậăắẳẵằặ",
        "d": "dđ",
        "e": "eéẻẽèẹêếểễềệ",
        "i": "iíỉĩìị",
        "o": "oóỏõòọôôốổỗồộơớởỡờợ",
        "u": "uúủũùụưứửữừự",
        "y": "yýỷỹỳỵ",
    }
    for k, v in ambiguous_chars.items():
        ambiguous_chars[k] = unicodedata.normalize("NFC", v)
    lowers = "".join(ambiguous_chars.values()) + string.ascii_lowercase
    uppers = lowers.upper()
    lowers = set(lowers)
    uppers = set(uppers)
    chars = lowers.union(uppers)
    puncs = ",.!?_()[]-'\": \t\n"
    puncs = unicodedata.normalize("NFC", puncs)
    puncs = set(puncs)

    vowels = set()
    for c in "aeiouy":
        vowels.update(ambiguous_chars[c])
    consonants = "bcdđghklmnprstvx" + "fjwz" # ph, th?
    consonants = set(consonants)

    @staticmethod
    def is_foreign(string):
        string = string.strip()
        if string == "":
            print("empty string!!!")
            return False

        string = unicodedata.normalize("NFC", string)
        char_set = set(string)
        intersection = char_set.intersection(Vietnamese.chars | Vietnamese.puncs)
        return len(intersection) < len(char_set)


def rm_diacritic(s):
    return unidecode.unidecode(s)


def normalize(s, form:str="NFD"):
    return unicodedata.normalize(form ,s)


ansi_code = {
    #"red":    "\033[91m",
    "red":    "\x1b[91m",
    "green":  "\033[92m",
    "yellow": "\033[93m",
    "blue":   "\033[94m",
    "pink":   "\033[95m",
    "teal":   "\033[96m",
    "grey":   "\033[97m",
    #"reset":    "\x1b[0m",
    "reset":    "\033[0m",
}


def roberta_infer_viz(
    datasets,
    *,
    topk: int = 5,
    device=torch.device("cpu"),
    tokenizer,
    model,
    data_collator,
):
    if "test" in datasets:
        ds = datasets["test"]
    else:
        ds = datasets["train"]
    sample = ds.shuffle().select([0])[0]

    collated = data_collator([sample])
    collated = {k: v.to(device) for k, v in collated.items()}
    model.to(device)

    masked_text = tokenizer.decode(collated["input_ids"][0])
    mask_token_indices = torch.where(collated["input_ids"] == tokenizer.mask_token_id)[1]

    labels = collated["labels"][0]
    gt_ids = sample["input_ids"][:]
    #for i, label in enumerate(labels):
    #    if label != -100:
    #        gt_ids[i] = label
    #gt_tokens = tokenizer.convert_ids_to_tokens(
    #    gt_ids, skip_special_tokens=False,
    #)
    #for i, (label, token) in enumerate(zip(labels, gt_tokens)):
    #    if i in mask_token_indices:
    #        if token[0] == 'Ġ':
    #            token = f' {ansi_code["green"]}{token[1:]}{ansi_code["reset"]}'
    #        else:
    #            token = f'{ansi_code["green"]}{token}{ansi_code["reset"]}'
    #    else:
    #        if token[0] == 'Ġ':
    #            token = f' {token[1:]}'
    #        else:
    #            continue
    #    gt_tokens[i] = token

    #gt_text = "".join(
    #    t for t in gt_tokens if t not in tokenizer.special_tokens_map.values())


    token_logits = model(**collated).logits
    mask_token_logits = token_logits[0, mask_token_indices, :]
    # Pick the <mask> candidates with the highest logits
    top_k_token_ids = torch.topk(mask_token_logits, topk, dim=1).indices.numpy()

    #print("(masked)")
    #print(f'"{masked_text}"', end="\n\n")

    #print("(gt)")
    #print(f'"{gt_text}"', end="\n\n")

    for kk in range(topk):
        pred_ids = collated["input_ids"][0].numpy()
        infer_token_ids = top_k_token_ids[:, kk]
        pred_ids[mask_token_indices] = infer_token_ids
        pred_tokens = tokenizer.convert_ids_to_tokens(
            pred_ids, skip_special_tokens=False,)
        for i, token in enumerate(pred_tokens):
            if i in mask_token_indices:
                gt_token_id = sample["labels"][i]
                gt_token = tokenizer.convert_ids_to_tokens(gt_token_id)
                if token[0] == 'Ġ':
                    token = f' {ansi_code["red"]}{token[1:]}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]}'
                else:
                    token = f'{ansi_code["red"]}{token}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]}'
            else:
                if token[0] == 'Ġ':
                    token = f'{token[1:]}'
                else:
                    if token in tokenizer.special_tokens_map.values():
                        token = ""
                    elif token in string.punctuation:
                        pass
                    else:
                        token = f' {token}'
            pred_tokens[i] = token

        #for i, (label, token) in enumerate(zip(labels, pred_tokens)):
        #    if i in mask_token_indices:
        #        if token[0] == 'Ġ':
        #            token = f' {ansi_code["red"]}{token[1:]}{ansi_code["reset"]}'
        #        else:
        #            token = f'{ansi_code["red"]}{token}{ansi_code["reset"]}'
        #    else:
        #        if token[0] == 'Ġ':
        #            token = f' {token[1:]}'
        #        else:
        #            pass
        #    pred_tokens[i] = token

        pred_text = "".join(t for t in pred_tokens)
        print(f'(pred {kk})')
        print(f'"{pred_text}"')
        print()


def bert_infer_viz(
    datasets,
    *,
    topk: int = 5,
    device=torch.device("cpu"),
    tokenizer,
    model,
    data_collator,
):
    """
    args
        data_collator, aka collate_fn
    """
    if isinstance(datasets, DatasetDict):
        if "test" in datasets:
            ds = datasets["test"]
        else:
            ds = datasets["train"]
    elif isinstance(datasets, Dataset):
        ds = datasets
    sample = ds.shuffle().select([0])[0]
    #print(f'tokenizer.decode(sample["labels"]) = "{tokenizer.decode(sample["labels"])}"')
    collated = data_collator([sample])
    # Put on the same assigned device
    collated = {k: v.to(device) for k, v in collated.items()}
    masked_text = tokenizer.decode(collated["input_ids"][0])
    mask_token_indices = torch.where(collated["input_ids"] == tokenizer.mask_token_id)[1]

    labels = sample["labels"]
    gt_ids = sample["input_ids"][:]
    #for i, label in enumerate(labels):
    #    if label != -100:
    #        gt_ids[i] = label
    #gt_tokens = tokenizer.convert_ids_to_tokens(
    #    gt_ids, skip_special_tokens=False,
    #)
    #print(f'{gt_tokens = }')
    #for i in mask_token_indices:
    #    token = gt_tokens[i]
    #    if token[:2] == '##':
    #        token = f'{ansi_code["green"]}{token[2:]}{ansi_code["reset"]}'
    #    else:
    #        token = f' {ansi_code["green"]}{token}{ansi_code["reset"]}'
    #    gt_tokens[i] = token
    #
    #gt_text = " ".join(
    #    t for t in gt_tokens if t not in tokenizer.special_tokens_map.values())

    model.to(device)
    token_logits = model(**collated).logits
    mask_token_logits = token_logits[0, mask_token_indices, :]
    # Pick the <mask> candidates with the highest logits
    #top_k_token_ids = torch.topk(mask_token_logits, topk, dim=1).indices.numpy()
    top_k_token_ids = torch.topk(mask_token_logits, topk, dim=1).indices.cpu().numpy()
    #top_k_token_ids = torch.topk(mask_token_logits, topk, dim=1).indices


    #print("(masked)")
    #print(f'"{masked_text}"', end="\n\n")

    #print("(gt)")
    #print(f'"{gt_text}"', end="\n\n")

    for kk in range(topk):
        pred_ids = collated["input_ids"][0].cpu().numpy()
        infer_token_ids = top_k_token_ids[:, kk]
        pred_ids[mask_token_indices] = infer_token_ids
        pred_tokens = tokenizer.convert_ids_to_tokens(
            pred_ids, skip_special_tokens=False,)
        #print(f'{pred_tokens = }')
        for i, token in enumerate(pred_tokens):
            if i in mask_token_indices:
                gt_token_id = sample["labels"][i]
                gt_token = tokenizer.convert_ids_to_tokens(gt_token_id)
                if token[:2] == '##':
                    token = f'{ansi_code["red"]}{token[2:]}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]}'
                else:
                    token = f' {ansi_code["red"]}{token}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]}'
            else:
                if token[:2] == '##':
                    token = f'{token[2:]}'
                else:
                    if token in tokenizer.special_tokens_map.values():
                        token = ""
                    elif token in string.punctuation:
                        pass
                    else:
                        token = f' {token}'
            pred_tokens[i] = token

        pred_text = "".join(t for t in pred_tokens)
        print(f'(pred {kk})')
        print(pred_text)
        print()


def phobert_infer_viz(
    datasets,
    *,
    topk: int = 5,
    device=torch.device("cpu"),
    tokenizer,
    model,
    data_collator,
):
    if "test" in datasets:
        ds = datasets["test"]
    else:
        ds = datasets["train"]
    sample = ds.shuffle().select([0])[0]
    collated = data_collator([sample])
    # Put on the same assigned device
    collated = {k: v.to(device) for k, v in collated.items()}
    masked_text = tokenizer.decode(collated["input_ids"][0])
    mask_token_indices = torch.where(collated["input_ids"] == tokenizer.mask_token_id)[1]
    #collated["input_ids"][0, mask_token_indices] = tokenizer.mask_token_id

    #labels = sample["labels"]
    #print(f'{tokenizer.decode(sample["labels"]) = }')
    #gt_ids = sample["input_ids"][:]

    model.to(device)
    token_logits = model(**collated).logits
    mask_token_logits = token_logits[0, mask_token_indices, :]
    # Pick the <mask> candidates with the highest logits
    top_k_token_ids = torch.topk(mask_token_logits, topk, dim=1).indices.numpy()

    #print("(masked)")
    #print(f'"{masked_text}"', end="\n\n")

    print("(gt)")
    print(f'"{tokenizer.decode(sample["labels"])}"', end="\n\n")

    for kk in range(topk):
        pred_ids = collated["input_ids"][0].numpy()
        infer_token_ids = top_k_token_ids[:, kk]
        pred_ids[mask_token_indices] = infer_token_ids
        pred_tokens = tokenizer.convert_ids_to_tokens(
            pred_ids, skip_special_tokens=False,)
        #print(f'{pred_tokens = }')
        for i, token in enumerate(pred_tokens):
            if i in mask_token_indices:
                gt_token_id = sample["labels"][i]
                gt_token = tokenizer.convert_ids_to_tokens(gt_token_id)
                if token[-2:] == '@@':
                    token = f'{ansi_code["red"]}{token[:-2]}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]}'
                else:
                    token = f'{ansi_code["red"]}{token}{ansi_code["reset"]}{ansi_code["green"]}{gt_token}{ansi_code["reset"]} '
            else:
                if token[-2:] == '@@':
                    token = f'{token[:-2]}'
                else:
                    if token in tokenizer.special_tokens_map.values():
                        token = ""
                    elif token in string.punctuation:
                        token = f'{token} '
                    else:
                        token = f'{token} '
            pred_tokens[i] = token

        pred_text = "".join(t for t in pred_tokens)
        print(f'(pred {kk})')
        print(pred_text)
        print()
